/**
 * Copyright(c) Live2D Inc. All rights reserved.
 *
 * Use of this source code is governed by the Live2D Open Software license
 * that can be found at https://www.live2d.com/eula/live2d-open-software-license-agreement_en.html.
 */

#include "LAppDelegate.hpp"
#include <iostream>
#include <sstream>
#include <kernel.h>
#include <libsysmodule.h>
#include <libdbg.h>
#include <gxm.h>
#include <touch.h>
#include <display.h>

#include "LAppGXM.hpp"
#include "LAppView.hpp"
#include "LAppPal.hpp"
#include "LAppDefine.hpp"
#include "LAppLive2DManager.hpp"
#include "LAppTextureManager.hpp"
#include <Rendering/GXM/CubismRenderer_GXM.hpp>

using namespace Csm;
using namespace std;
using namespace LAppDefine;

namespace {
    LAppDelegate* s_instance = NULL;
}

LAppDelegate* LAppDelegate::GetInstance()
{
    if (s_instance == NULL)
    {
        s_instance = new LAppDelegate();
    }

    return s_instance;
}

void LAppDelegate::ReleaseInstance()
{
    if (s_instance != NULL)
    {
        delete s_instance;
    }

    s_instance = NULL;
}



bool LAppDelegate::Initialize()
{
    if (DebugLogEnable)
    {
        LAppPal::PrintLog("START");
    }

    sceDbgSetMinimumLogLevel(SCE_DBG_LOG_LEVEL_TRACE);

    sceTouchSetSamplingState(SCE_TOUCH_PORT_FRONT, SCE_TOUCH_SAMPLING_STATE_START);

    LAppGXM::InitParams gxmContextParams;
    sceClibMemset(&gxmContextParams, 0, sizeof(LAppGXM::InitParams));
    gxmContextParams.msaa = (SceGxmMultisampleMode)MsaaMode;
    LAppGXM *gxmContext = new LAppGXM(&gxmContextParams);

    // ウィンドウサイズ記憶
    _windowWidth = DisplayWidth;
    _windowHeight = DisplayHeight;

    //AppViewの初期化
    _view->Initialize();

    // Cubism SDK の初期化
    InitializeCubism();

    SetRootDirectory();

    //load model
    LAppLive2DManager::GetInstance();

    //load sprite
    _view->InitializeSprite();

    return 1;
}

void LAppDelegate::Release()
{
    LAppGXM::ReleaseInstance();

    delete _textureManager;
    delete _view;

    // リソースを解放
    LAppLive2DManager::ReleaseInstance();

    //Cubism SDK の解放
    CubismFramework::Dispose();
}

void LAppDelegate::Run()
{
    //メインループ
    while (!_isEnd)
    {
        CheckControls();

        // 時間更新
        LAppPal::UpdateTime();

        //描画更新
        _view->Render();

        // バッファの入れ替え
        LAppGXM::GetInstance()->EndRendering();
    }

    Release();

    LAppDelegate::ReleaseInstance();
}

LAppDelegate::LAppDelegate():
    _cubismOption(),
    _captured(false),
    _mouseX(0.0f),
    _mouseY(0.0f),
    _isEnd(false),
    _windowWidth(0),
    _windowHeight(0)
{
    _rootDirectory = "";
    _view = new LAppView();
    _textureManager = new LAppTextureManager();
}

LAppDelegate::~LAppDelegate()
{

}

void LAppDelegate::InitializeCubism()
{
    //setup cubism
    _cubismOption.LogFunction = LAppPal::PrintMessage;
    _cubismOption.LoggingLevel = LAppDefine::CubismLoggingLevel;
    Csm::CubismFramework::StartUp(&_cubismAllocator, &_cubismOption);

    //Initialize cubism
    CubismFramework::Initialize();

    //default proj
    CubismMatrix44 projection;

    LAppPal::UpdateTime();
}

void LAppDelegate::CheckControls()
{
    SceTouchData td;
    int action = 0;

    sceTouchPeek(SCE_TOUCH_PORT_FRONT, &td, 1);

    _mouseX = (float)td.report->x / 2.0f;
    _mouseY = (float)td.report->y / 2.0f;

    if (td.reportNum != _tdOld.reportNum && td.reportNum == 1)
        action = 1;
    else if (td.reportNum != _tdOld.reportNum && td.reportNum == 0) {
        action = 2;
        _mouseX = (float)_tdOld.report->x / 2.0f;
        _mouseY = (float)_tdOld.report->y / 2.0f;
    }

    if (1 == action)
    {
        _captured = true;
        _view->OnTouchesBegan(_mouseX, _mouseY);
    }
    else if (2 == action)
    {
        if (_captured)
        {
            _captured = false;
            _view->OnTouchesEnded(_mouseX, _mouseY);
        }
    }

    if (_captured && td.reportNum != 0)
    {
        _view->OnTouchesMoved(_mouseX, _mouseY);
    }

    _tdOld = td;
}

void LAppDelegate::CreateShader(LAppView *view)
{
    LAppGXM *gxmContext = LAppGXM::GetInstance();

    SceGxmShaderPatcherId vertexProgramId;
    SceGxmShaderPatcherId fragmentProgramId;

    int err = sceGxmShaderPatcherRegisterProgram(gxmContext->shaderPatcher, vertexShader, &vertexProgramId);
    if (err != SCE_OK) {
        CubismLogError("vertexShader sceGxmShaderPatcherRegisterProgram(): 0x%X", err);
        return;
    }

    err = sceGxmShaderPatcherRegisterProgram(gxmContext->shaderPatcher, fragmentShader, &fragmentProgramId);
    if (err != SCE_OK) {
        CubismLogError("fragmentShader sceGxmShaderPatcherRegisterProgram(): 0x%X", err);
        return;
    }

    SceGxmBlendInfo blendInfo;
    blendInfo.colorMask = SCE_GXM_COLOR_MASK_ALL;
    blendInfo.alphaFunc = SCE_GXM_BLEND_FUNC_ADD;
    blendInfo.colorFunc = SCE_GXM_BLEND_FUNC_ADD;
    blendInfo.alphaDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
    blendInfo.colorDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
    blendInfo.alphaSrc = SCE_GXM_BLEND_FACTOR_SRC_ALPHA;
    blendInfo.colorSrc = SCE_GXM_BLEND_FACTOR_SRC_ALPHA;

    const SceGxmProgramParameter *paramPositionAttribute = sceGxmProgramFindParameterByName(vertexShader, "position");
    const SceGxmProgramParameter *paramUvAttribute = sceGxmProgramFindParameterByName(vertexShader, "uv");

    SceGxmVertexAttribute vertexAttributes[2];
    SceGxmVertexStream vertexStreams[1];
    vertexAttributes[0].streamIndex = 0;
    vertexAttributes[0].offset = 0;
    vertexAttributes[0].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
    vertexAttributes[0].componentCount = 2;
    vertexAttributes[0].regIndex = sceGxmProgramParameterGetResourceIndex(paramUvAttribute);
    vertexAttributes[1].streamIndex = 0;
    vertexAttributes[1].offset = 8;
    vertexAttributes[1].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
    vertexAttributes[1].componentCount = 2;
    vertexAttributes[1].regIndex = sceGxmProgramParameterGetResourceIndex(paramPositionAttribute);
    vertexStreams[0].stride = 16;
    vertexStreams[0].indexSource = SCE_GXM_INDEX_SOURCE_INDEX_16BIT;

    err = sceGxmShaderPatcherCreateVertexProgram(
        gxmContext->shaderPatcher,
        vertexProgramId,
        vertexAttributes,
        2,
        vertexStreams,
        1,
        &view->vertexProgram);

    if (err != SCE_OK) {
        CubismLogError("color sceGxmShaderPatcherCreateVertexProgram(): 0x%X", err);
        return;
    }

    err = sceGxmShaderPatcherCreateFragmentProgram(
        gxmContext->shaderPatcher,
        fragmentProgramId,
        SCE_GXM_OUTPUT_REGISTER_FORMAT_UCHAR4,
        gxmContext->GetMsaa(),
        &blendInfo,
        vertexShader,
        &view->fragmentProgram);

    if (err != SCE_OK) {
        CubismLogError("color sceGxmShaderPatcherCreateVertexProgram(): 0x%X", err);
        return;
    }
}

void LAppDelegate::SetRootDirectory()
{
    this->_rootDirectory = "app0:";
}

Csm::csmVector<string> LAppDelegate::Split(const std::string& baseString, char delimiter)
{
    Csm::csmVector<string> elems;
    stringstream ss(baseString);
    string item;
    while(getline(ss, item, delimiter))
    {
        if(!item.empty())
        {
            elems.PushBack(item);
        }
    }
    return elems;
}
