
#include "LAppGXM.hpp"
#include "LAppDefine.hpp"
#include "LAppAllocator.hpp"

#include <CubismModelSettingJson.hpp>
#include <Motion/CubismMotion.hpp>
#include <Physics/CubismPhysics.hpp>
#include <CubismDefaultParameterId.hpp>
#include <Rendering/GXM/CubismRenderer_GXM.hpp>
#include <Utils/CubismString.hpp>
#include <Id/CubismIdManager.hpp>
#include <Motion/CubismMotionQueueEntry.hpp>

#include <kernel.h>
#include <display.h>
#include <gxm.h>
#include <gxt.h>

using namespace LAppDefine;

namespace {
	LAppGXM *s_instance = NULL;
}

LAppGXM *LAppGXM::GetInstance()
{
	return s_instance;
}

void LAppGXM::ReleaseInstance()
{
	if (s_instance != NULL)
	{
		delete s_instance;
	}

	s_instance = NULL;
}

LAppGXM::LAppGXM(InitParams *params)
	: msaa(SCE_GXM_MULTISAMPLE_NONE)
	, _bufferIndex(0)
{
	int err;

	if (params == NULL) {
		return;
	}

	if (!params->paramBufferSize)
		params->paramBufferSize = SCE_GXM_DEFAULT_PARAMETER_BUFFER_SIZE;
	if (!params->vdmRingBufferSize)
		params->vdmRingBufferSize = SCE_GXM_DEFAULT_VDM_RING_BUFFER_SIZE;
	if (!params->vertexRingBufferSize)
		params->vertexRingBufferSize = SCE_GXM_DEFAULT_VERTEX_RING_BUFFER_SIZE;
	if (!params->fragmentRingBufferSize)
		params->fragmentRingBufferSize = SCE_GXM_DEFAULT_FRAGMENT_RING_BUFFER_SIZE;
	if (!params->fragmentUsseRingBufferSize)
		params->fragmentUsseRingBufferSize = SCE_GXM_DEFAULT_FRAGMENT_USSE_RING_BUFFER_SIZE;

	SceGxmInitializeParams initializeParams;
	sceClibMemset(&initializeParams, 0, sizeof(SceGxmInitializeParams));
	initializeParams.flags = 0;
	initializeParams.displayQueueMaxPendingCount = 2;
	initializeParams.displayQueueCallback = LAppGXM::DisplayCallback;
	initializeParams.displayQueueCallbackDataSize = sizeof(DisplayCallbackArg);
	initializeParams.parameterBufferSize = params->paramBufferSize;

	err = sceGxmInitialize(&initializeParams);

	if (err != SCE_OK) {
		CubismLogError("sceGxmInitialize(): 0x%X", err);
		return;
	}

	_validRegion.xMax = DisplayWidth - 1;
	_validRegion.yMax = DisplayHeight - 1;

	msaa = params->msaa;

	// allocate ring buffer memory

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_USER_NC,
		SCE_GXM_MEMORY_ATTRIB_READ,
		params->vdmRingBufferSize,
		4,
		&_vdmRingBufferMem);

	if (err != SCE_OK) {
		CubismLogError("VDM ring buffer allocation failed: 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_USER_NC,
		SCE_GXM_MEMORY_ATTRIB_READ,
		params->vertexRingBufferSize,
		4,
		&_vertexRingBufferMem);

	if (err != SCE_OK) {
		CubismLogError("Vertex ring buffer allocation failed: 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_USER_NC,
		SCE_GXM_MEMORY_ATTRIB_READ,
		params->fragmentRingBufferSize,
		4,
		&_fragmentRingBufferMem);

	if (err != SCE_OK) {
		CubismLogError("Fragment ring buffer allocation failed: 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_FRAGMENT_USSE,
		SCE_GXM_MEMORY_ATTRIB_READ,
		params->fragmentUsseRingBufferSize,
		4096,
		&_fragmentUsseRingBufferMem);

	if (err != SCE_OK) {
		CubismLogError("Fragment USSE ring buffer allocation failed: 0x%X", err);
		return;
	}

	SceGxmContextParams contextParams;
	sceClibMemset(&contextParams, 0, sizeof(SceGxmContextParams));
	contextParams.hostMem = malloc(SCE_GXM_MINIMUM_CONTEXT_HOST_MEM_SIZE);
	contextParams.hostMemSize = SCE_GXM_MINIMUM_CONTEXT_HOST_MEM_SIZE;
	contextParams.vdmRingBufferMem = _vdmRingBufferMem->mappedBase;
	contextParams.vdmRingBufferMemSize = params->vdmRingBufferSize;
	contextParams.vertexRingBufferMem = _vertexRingBufferMem->mappedBase;
	contextParams.vertexRingBufferMemSize = params->vertexRingBufferSize;
	contextParams.fragmentRingBufferMem = _fragmentRingBufferMem->mappedBase;
	contextParams.fragmentRingBufferMemSize = params->fragmentRingBufferSize;
	contextParams.fragmentUsseRingBufferMem = _fragmentUsseRingBufferMem->mappedBase;
	contextParams.fragmentUsseRingBufferMemSize = params->fragmentUsseRingBufferSize;
	contextParams.fragmentUsseRingBufferOffset = _fragmentUsseRingBufferMem->offset;

	err = sceGxmCreateContext(&contextParams, &immContext);

	if (err != SCE_OK) {
		CubismLogError("sceGxmCreateContext(): 0x%X", err);
		return;
	}

	// set up parameters
	SceGxmRenderTargetParams renderTargetParams;
	sceClibMemset(&renderTargetParams, 0, sizeof(SceGxmRenderTargetParams));
	renderTargetParams.flags = 0;
	renderTargetParams.width = MaxDisplayWidth;
	renderTargetParams.height = MaxDisplayHeight;
	renderTargetParams.scenesPerFrame = 2;
	renderTargetParams.multisampleMode = msaa;
	renderTargetParams.multisampleLocations = 0;
	renderTargetParams.driverMemBlock = SCE_UID_INVALID_UID;

	// create the render target
	err = sceGxmCreateRenderTarget(&renderTargetParams, &renderTarget);

	if (err != SCE_OK) {
		CubismLogError("sceGxmCreateRenderTarget(): 0x%X", err);
		return;
	}

	// allocate memory and sync objects for display buffers
	for (int i = 0; i < 2; i++) {

		// allocate memory for display
		sceGxmAllocDeviceMemLinux(
			SCE_GXM_DEVICE_HEAP_ID_CDRAM,
			SCE_GXM_MEMORY_ATTRIB_READ | SCE_GXM_MEMORY_ATTRIB_WRITE,
			4 * DisplayStride * DisplayHeight,
			SCE_GXM_COLOR_SURFACE_ALIGNMENT,
			&_displayBufferMem[i]);

		displayBufferData[i] = _displayBufferMem[i]->mappedBase;

		// memset the buffer to black
		sceGxmTransferFill(
			0xFF000000,
			SCE_GXM_TRANSFER_FORMAT_U8U8U8U8_ABGR,
			displayBufferData[i],
			0,
			0,
			DisplayWidth,
			DisplayHeight,
			DisplayStride,
			NULL,
			0,
			NULL);

		// create a sync object that we will associate with this buffer
		err = sceGxmSyncObjectCreate(&displayBufferSync[i]);

		if (err != SCE_OK) {
			CubismLogError("PVRSRVAllocSyncInfo(): 0x%X", err);
			return;
		}

		// initialize a color surface for this display buffer
		err = sceGxmColorSurfaceInit(
			&displayColorSurface[i],
			SCE_GXM_COLOR_FORMAT_A8B8G8R8,
			SCE_GXM_COLOR_SURFACE_LINEAR,
			(msaa == SCE_GXM_MULTISAMPLE_NONE) ? SCE_GXM_COLOR_SURFACE_SCALE_NONE : SCE_GXM_COLOR_SURFACE_SCALE_MSAA_DOWNSCALE,
			SCE_GXM_OUTPUT_REGISTER_SIZE_32BIT,
			DisplayWidth,
			DisplayHeight,
			DisplayStride,
			displayBufferData[i]);

		if (err != SCE_OK) {
			CubismLogError("sceGxmColorSurfaceInit(): 0x%X", err);
			return;
		}
	}

	// compute the memory footprint of the depth buffer
	const uint32_t alignedWidth = ALIGN(DisplayWidth, SCE_GXM_TILE_SIZEX);
	const uint32_t alignedHeight = ALIGN(DisplayHeight, SCE_GXM_TILE_SIZEY);
	uint32_t sampleCount = alignedWidth * alignedHeight;
	uint32_t depthStrideInSamples = alignedWidth;
	if (msaa == SCE_GXM_MULTISAMPLE_4X) {
		// samples increase in X and Y
		sampleCount *= 4;
		depthStrideInSamples *= 2;
	}
	else if (msaa == SCE_GXM_MULTISAMPLE_2X) {
		// samples increase in Y only
		sampleCount *= 2;
	}

	// allocate the depth buffer
	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_CDRAM,
		SCE_GXM_MEMORY_ATTRIB_READ | SCE_GXM_MEMORY_ATTRIB_WRITE,
		4 * sampleCount,
		SCE_GXM_DEPTHSTENCIL_SURFACE_ALIGNMENT,
		&_depthBufferMem);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}

	// allocate the stencil buffer
	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_CDRAM,
		SCE_GXM_MEMORY_ATTRIB_READ | SCE_GXM_MEMORY_ATTRIB_WRITE,
		4 * sampleCount,
		SCE_GXM_DEPTHSTENCIL_SURFACE_ALIGNMENT,
		&_stencilBufferMem);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}

	// create the SceGxmDepthStencilSurface structure
	err = sceGxmDepthStencilSurfaceInit(
		&depthStencilSurface,
		SCE_GXM_DEPTH_STENCIL_FORMAT_S8D24,
		SCE_GXM_DEPTH_STENCIL_SURFACE_TILED,
		depthStrideInSamples,
		_depthBufferMem->mappedBase,
		_stencilBufferMem->mappedBase);

	if (err != SCE_OK) {
		CubismLogError("sceGxmDepthStencilSurfaceInit(): 0x%X", err);
		return;
	}

	// set the stencil test reference (this is currently assumed to always remain 1 after here for region clipping)
	sceGxmSetFrontStencilRef(immContext, 1);

	// set the stencil function (this wouldn't actually be needed, as the set clip rectangle function has to call this at the begginning of every scene)
	sceGxmSetFrontStencilFunc(
		immContext,
		SCE_GXM_STENCIL_FUNC_ALWAYS,
		SCE_GXM_STENCIL_OP_KEEP,
		SCE_GXM_STENCIL_OP_KEEP,
		SCE_GXM_STENCIL_OP_KEEP,
		0xFF,
		0xFF);

	// set buffer sizes for this sample
	const uint32_t patcherBufferSize = 64 * 1024;
	const uint32_t patcherVertexUsseSize = 64 * 1024;
	const uint32_t patcherFragmentUsseSize = 64 * 1024;

	// allocate memory for buffers and USSE code
	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_USER_NC,
		SCE_GXM_MEMORY_ATTRIB_READ | SCE_GXM_MEMORY_ATTRIB_WRITE,
		patcherBufferSize,
		4,
		&_patcherBufferMem);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_VERTEX_USSE,
		SCE_GXM_MEMORY_ATTRIB_READ,
		patcherVertexUsseSize,
		4096,
		&_patcherVertexUsseMem);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_FRAGMENT_USSE,
		SCE_GXM_MEMORY_ATTRIB_READ,
		patcherFragmentUsseSize,
		4096,
		&_patcherFragmentUsseMem);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}

	// create a shader patcher
	SceGxmShaderPatcherParams patcherParams;
	sceClibMemset(&patcherParams, 0, sizeof(SceGxmShaderPatcherParams));
	patcherParams.userData = NULL;
	patcherParams.hostAllocCallback = &PatcherHostAlloc;
	patcherParams.hostFreeCallback = &PatcherHostFree;
	patcherParams.bufferAllocCallback = NULL;
	patcherParams.bufferFreeCallback = NULL;
	patcherParams.bufferMem = _patcherBufferMem->mappedBase;
	patcherParams.bufferMemSize = _patcherBufferMem->size;
	patcherParams.vertexUsseAllocCallback = NULL;
	patcherParams.vertexUsseFreeCallback = NULL;
	patcherParams.vertexUsseMem = _patcherVertexUsseMem->mappedBase;
	patcherParams.vertexUsseMemSize = _patcherVertexUsseMem->size;
	patcherParams.vertexUsseOffset = _patcherVertexUsseMem->offset;
	patcherParams.fragmentUsseAllocCallback = NULL;
	patcherParams.fragmentUsseFreeCallback = NULL;
	patcherParams.fragmentUsseMem = _patcherFragmentUsseMem->mappedBase;
	patcherParams.fragmentUsseMemSize = _patcherFragmentUsseMem->size;
	patcherParams.fragmentUsseOffset = _patcherFragmentUsseMem->offset;

	err = sceGxmShaderPatcherCreate(&patcherParams, &shaderPatcher);

	if (err != SCE_OK) {
		CubismLogError("sceGxmShaderPatcherCreate(): 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_USER_NC,
		SCE_GXM_MEMORY_ATTRIB_READ,
		4 * 1024,
		4096,
		&vertexStorage);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}

	err = sceGxmAllocDeviceMemLinux(
		SCE_GXM_DEVICE_HEAP_ID_USER_NC,
		SCE_GXM_MEMORY_ATTRIB_READ,
		4 * 1024,
		4096,
		&indexStorage);

	if (err != SCE_OK) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", err);
		return;
	}


	uint16_t *indexStream = (uint16_t *)(indexStorage->mappedBase);

	indexStream[0] = 0;
	indexStream[1] = 1;
	indexStream[2] = 2;
	indexStream[3] = 3;
	indexStream[4] = 4;

	sceGxmSetViewportEnable(immContext, SCE_GXM_VIEWPORT_ENABLED);

	s_instance = this;
}

LAppGXM::~LAppGXM()
{
	CubismLogWarning("WARNING! GXM termination is not implemented");
}

void LAppGXM::DisplayCallback(const void *callbackData)
{
	const DisplayCallbackArg* arg = (const DisplayCallbackArg*)callbackData;
	SceDisplayFrameBuf framebuf;

	sceClibMemset(&framebuf, 0, sizeof(SceDisplayFrameBuf));
	framebuf.size = sizeof(SceDisplayFrameBuf);
	framebuf.base = arg->address;
	framebuf.pitch = DisplayStride;
	framebuf.pixelformat = SCE_DISPLAY_PIXELFORMAT_A8B8G8R8;
	framebuf.width = DisplayWidth;
	framebuf.height = DisplayHeight;
	sceDisplaySetFrameBuf(&framebuf, SCE_DISPLAY_UPDATETIMING_NEXTVSYNC);

	if (SwapInterval) {
		sceDisplayWaitVblankStartMulti(SwapInterval);
	}
}

void *LAppGXM::PatcherHostAlloc(void *user_data, uint32_t size)
{
	void* pMem = malloc(size);
	if (pMem == NULL)
		CubismLogError("patcherHostAlloc(): malloc() returned NULL");
	return pMem;
}

void LAppGXM::PatcherHostFree(void *user_data, void *mem)
{
	free(mem);
}

void LAppGXM::StartScene()
{
	sceGxmBeginScene(
		immContext,
		0,
		renderTarget,
		&_validRegion,
		NULL,
		displayBufferSync[_bufferIndex],
		&displayColorSurface[_bufferIndex],
		&depthStencilSurface);
}

void LAppGXM::EndScene()
{
	sceGxmEndScene(immContext, NULL, NULL);
}

void LAppGXM::EndRendering()
{
	sceGxmPadHeartbeat(&displayColorSurface[_bufferIndex], displayBufferSync[_bufferIndex]);

	int oldFb = (_bufferIndex + 1) % 2;

	DisplayCallbackArg displayData;
	displayData.address = displayBufferData[_bufferIndex];
	sceGxmDisplayQueueAddEntry(
		displayBufferSync[oldFb],
		displayBufferSync[_bufferIndex],
		&displayData);

	_bufferIndex = (_bufferIndex + 1) % 2;
}

LAppGXM::GXMTexture *LAppGXM::LoadGxtTexture(std::string fileName, int index)
{
	int ret;

	GXMTexture *tex = new GXMTexture();

	const char *path = fileName.c_str();
	SceIoStat file_stat;
	sceIoGetstat(path, &file_stat);

	ret = sceGxmAllocDeviceMemLinux(SCE_GXM_DEVICE_HEAP_ID_USER_NC, SCE_GXM_MEMORY_ATTRIB_READ, (uint32_t)file_stat.st_size, 4096, &tex->memData);
	if (ret < 0) {
		CubismLogError("sceGxmAllocDeviceMemLinux(): 0x%X", ret);
		free(tex);
		return NULL;
	}

	unsigned int read_size;

	SceUID fd = sceIoOpen(path, SCE_O_RDONLY, 0);

	if (fd < 0)
		CubismLogError("Can't open file %s sceIoOpen(): 0x%X", path, ret);

	read_size = sceIoRead(fd, tex->memData->mappedBase, (SceSize)file_stat.st_size);
	sceIoClose(fd);

	if (read_size != file_stat.st_size) {
		CubismLogError("Can't read file %s", path);
	}

	ret = sceGxtCheckData(tex->memData->mappedBase);

	if (ret < 0) {
		CubismLogError("sceGxtCheckData(): 0x%X", ret);
	}

	const void *actual_texture_data = sceGxtGetDataAddress(tex->memData->mappedBase);

	ret = sceGxtInitTexture(&tex->gxmTexture, tex->memData->mappedBase, actual_texture_data, index);

	if (ret < 0) {
		CubismLogError("sceGxtInitTexture(): 0x%X", ret);
	}

	return tex;
}

void LAppGXM::UnloadGxtTexture(LAppGXM::GXMTexture *texture)
{
	if (texture) {
		sceGxmFreeDeviceMemLinux(texture->memData);
		delete texture;
	}
}

csmInt32 LAppGXM::GetDisplayIndex()
{
	return _bufferIndex;
}

void LAppGXM::WaitRenderingDone()
{
	sceGxmFinish(immContext);
}