/**
 * Copyright(c) Live2D Inc. All rights reserved.
 *
 * Use of this source code is governed by the Live2D Open Software license
 * that can be found at https://www.live2d.com/eula/live2d-open-software-license-agreement_en.html.
 */


#include <kernel.h>
#include <gxm.h>
#include "LAppSprite.hpp"
#include "LAppDelegate.hpp"
#include "LAppDefine.hpp"
#include "LAppGXM.hpp"
#include "LAppView.hpp"

using namespace LAppDefine;

namespace {
    uint32_t spriteVertexStreamOffset = 0;
}

LAppSprite::LAppSprite(float x, float y, float width, float height, LAppGXM::GXMTexture *texture)
    : _rect()
{
    _rect.left = (x - width * 0.5f);
    _rect.right = (x + width * 0.5f);
    _rect.up = (y + height * 0.5f);
    _rect.down = (y - height * 0.5f);
    _texture = texture;
    _vertexStreamOffset = spriteVertexStreamOffset;
    spriteVertexStreamOffset += 16;

    _colorLocation = sceGxmProgramFindParameterByName(fragmentShader, "baseColor");

    _spriteColor[0] = 1.0f;
    _spriteColor[1] = 1.0f;
    _spriteColor[2] = 1.0f;
    _spriteColor[3] = 1.0f;
}

LAppSprite::~LAppSprite()
{
}

void LAppSprite::Render(LAppView *view) const
{
    LAppGXM *gxmContext = LAppGXM::GetInstance();

    float *vertexStream = (float *)(gxmContext->vertexStorage->mappedBase);
    vertexStream += _vertexStreamOffset;

    //uvVertex
    vertexStream[0] = 1.0f;
    vertexStream[1] = 0.0f;
    vertexStream[4] = 0.0f;
    vertexStream[5] = 0.0f;
    vertexStream[8] = 0.0f;
    vertexStream[9] = 1.0f;
    vertexStream[12] = 1.0f;
    vertexStream[13] = 1.0f;

    float cWidth = DisplayWidth * 0.5f;
    float cHeight = DisplayHeight * 0.5f;

    //positionVertex
    vertexStream[2] = (_rect.right - cWidth) / cWidth;
    vertexStream[3] = (_rect.up - cHeight) / cHeight;
    vertexStream[6] = (_rect.left - cWidth) / cWidth;
    vertexStream[7] = (_rect.up - cHeight) / cHeight;
    vertexStream[10] = (_rect.left - cWidth) / cWidth;
    vertexStream[11] = (_rect.down - cHeight) / cHeight;
    vertexStream[14] = (_rect.right - cWidth) / cWidth;
    vertexStream[15] = (_rect.down - cHeight) / cHeight;

    sceGxmSetVertexProgram(gxmContext->immContext, view->vertexProgram);
    sceGxmSetFragmentProgram(gxmContext->immContext, view->fragmentProgram);

    sceGxmSetVertexStream(gxmContext->immContext, 0, vertexStream);
    sceGxmSetFragmentTexture(gxmContext->immContext, 0, &_texture->gxmTexture);

    void *fragmentDefaultBuffer;
    sceGxmReserveFragmentDefaultUniformBuffer(gxmContext->immContext, &fragmentDefaultBuffer);
    sceGxmSetUniformDataF(fragmentDefaultBuffer, _colorLocation, 0, 4, _spriteColor);

    uint16_t *indexStream = (uint16_t *)(gxmContext->indexStorage->mappedBase);

    // モデルの描画
    sceGxmDraw(gxmContext->immContext, SCE_GXM_PRIMITIVE_TRIANGLE_FAN, SCE_GXM_INDEX_FORMAT_U16, indexStream, 4);
}

void LAppSprite::RenderImmidiate(LAppView *view, SceGxmTexture* texture, const csmFloat32 uvVertex[8]) const
{
    LAppGXM *gxmContext = LAppGXM::GetInstance();

    float *vertexStream = (float *)(gxmContext->vertexStorage->mappedBase);
    vertexStream += _vertexStreamOffset;

    //uvVertex
    vertexStream[0] = uvVertex[0];
    vertexStream[1] = uvVertex[1];
    vertexStream[4] = uvVertex[2];
    vertexStream[5] = uvVertex[3];
    vertexStream[8] = uvVertex[4];
    vertexStream[9] = uvVertex[5];
    vertexStream[12] = uvVertex[6];
    vertexStream[13] = uvVertex[7];

    float cWidth = DisplayWidth * 0.5f;
    float cHeight = DisplayHeight * 0.5f;

    //positionVertex
    vertexStream[2] = (_rect.right - cWidth) / cWidth;
    vertexStream[3] = (_rect.up - cHeight) / cHeight;
    vertexStream[6] = (_rect.left - cWidth) / cWidth;
    vertexStream[7] = (_rect.up - cHeight) / cHeight;
    vertexStream[10] = (_rect.left - cWidth) / cWidth;
    vertexStream[11] = (_rect.down - cHeight) / cHeight;
    vertexStream[14] = (_rect.right - cWidth) / cWidth;
    vertexStream[15] = (_rect.down - cHeight) / cHeight;

    sceGxmSetVertexProgram(gxmContext->immContext, view->vertexProgram);
    sceGxmSetFragmentProgram(gxmContext->immContext, view->fragmentProgram);

    sceGxmSetVertexStream(gxmContext->immContext, 0, vertexStream);
    sceGxmSetFragmentTexture(gxmContext->immContext, 0, texture);

    void *fragmentDefaultBuffer;
    sceGxmReserveFragmentDefaultUniformBuffer(gxmContext->immContext, &fragmentDefaultBuffer);
    sceGxmSetUniformDataF(fragmentDefaultBuffer, _colorLocation, 0, 4, _spriteColor);

    uint16_t *indexStream = (uint16_t *)(gxmContext->indexStorage->mappedBase);

    // モデルの描画
    sceGxmDraw(gxmContext->immContext, SCE_GXM_PRIMITIVE_TRIANGLE_FAN, SCE_GXM_INDEX_FORMAT_U16, indexStream, 4);
}

bool LAppSprite::IsHit(float pointX, float pointY) const
{
    //Y座標は変換する必要あり
    float y = DisplayHeight - pointY;

    return (pointX >= _rect.left && pointX <= _rect.right && y <= _rect.up && y >= _rect.down);
}

void LAppSprite::SetColor(float r, float g, float b, float a)
{
    _spriteColor[0] = r;
    _spriteColor[1] = g;
    _spriteColor[2] = b;
    _spriteColor[3] = a;
}

void LAppSprite::ResetRect(float x, float y, float width, float height)
{
    _rect.left = (x - width * 0.5f);
    _rect.right = (x + width * 0.5f);
    _rect.up = (y + height * 0.5f);
    _rect.down = (y - height * 0.5f);
}
